unit RomDis;

//---------------------------------------------------------------------------//
//    Copyright (c) 2008  Meindert Kuipers, Netherlands                      //
//    meindert@kuiprs.nl                  www.kuiprs.nl                      //
//                                                                           //
// This program is free software; you can redistribute it and/or             //
// modify it under the terms of the GNU General Public License               //
// as published by the Free Software Foundation; either version 2            //
// of the License, or (at your option) any later version.                    //
//                                                                           //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
// GNU General Public License for more details.                              //
//                                                                           //
// You should have received a copy of the GNU General Public License         //
// along with this program; if not, write to the Free Software               //
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA//
//---------------------------------------------------------------------------//
//
//---------------------------------------------------------------------------//
//  ROMDIS.PAS                                                               //
//  HP41 ROM Disassembler                                                    //
//  Ver  Date     Description                                                //
//  0.01 Mar 2007 First draft integrated in M2kM based on old version        //
//  1.10 Apr 2007 User Code Function names supported                         //
//                Message given for Function Name in other Page              //
//  1.11 Feb 2008 Will now directly disassemble usercode before label        //
//                Disassembly prints * before label                          //
//                Will not attempt to disasssemble function names            //
//                Empty line inserted before start of function               //
//                Explains meaning of first two words before User Code       //
//                Improved layout of user code disassembly                   //
//                Added MainFrame labels, XROM, local labels, 8K dis         //
//  1.20 Apr 2008 Fixed some bugs, added 8K assembly                         //
//                Added SkipNOPs,                                            //
//       May 2008 Added text comments to XEQ and LBL                         //
//  1.50 May 2008 Final release                                              //
//  1.51 Aug 2008 Added Label search for local JC/JNC and GOTO/GOSUB         //
//       Nov 2008 Added correct disassembly for ERROR, MSG, MSGA             //
//                                                                           //
//---------------------------------------------------------------------------//

interface

uses
  Globals, HP41_Globals, GlobalConstants, Procs, SysUtils, ListWindow, StrUtils;

var
  Mode, IClass, FirstWord, SecondWord, StartAdr, W, WW, A: word;
  St, Hulp: string;
  LabLen: integer;

function Fetch(MAR: word): word;
function GetInstr(Adr: word): word;
function GetID(I: word): word;
procedure SetID(ID, Adr: word);
function HPChar(W: word): char;
function FindLabel(Adr: word): string12;
procedure AddLabel(Adr: word; Lab: string);
procedure FindMainFrameFile(Adr: word; var MainLabel, Comment: string);
procedure FindXROMFile(I, J: Integer; var S1, S2: string);
procedure DisAssem(var Adr: word; I: word; var S: string);

implementation

function Fetch(MAR: word): word;
// Fetches a word, only from current page!
begin
  MAR := MAR and ROMEnd;
  Fetch := Page[MAR];
end; // Fetch

function GetInstr(Adr: word): word;
// gets the clean OpCode, without ID
begin
  GetInstr := Fetch(Adr and ROMEnd) and $03FF;
end;

function GetID(I: word): word;
// decodes the opcode ID from the instruction
begin
  GetID := (I and $FC00) shr 10;
end;

procedure ClearMCodeLabel(Adr: word);
// Removes the label if it starts with LB_
begin
  if Pos('LB_', LabelTable[Adr and ROMEnd]) = 1 then
    LabelTable[Adr and ROMEnd] := '          ';
end;

procedure SetID(ID, Adr: word);
// puts the ID code into the upper 6 bits of the instruction
begin
  Page[Adr and ROMEnd] := GetInstr(Adr) or (ID shl 10);
  if ID = UserProgram then
    ClearMCodeLabel(Adr);
end;

function HPChar(W: word): Char;
begin
  HPChar := HPCharS[(W and $003F) + 1];
end;

function FindLabel(Adr: word): String12;
// Locates a Label
var
  S: string;
  P: integer;
begin
  S := LabelTable[Adr and ROMEnd];
  if PrefMnem = 3 then begin
   S := '[' + S;
   P := Pos(Space, S);         // finds the first space in the string
   if P <> 0 then S[P] := ']';          // replace space with bracket
   S := S + Space;
  end else
    S := S + '  ';
  FindLabel := S;
end; // FindLabel


procedure AddLabel(Adr: word; Lab: string);
// Adds a label to the table if that position was empty, expands the string to 10 chars
// also checks for range arrors
var
  i, L: integer;
  S: String10;
begin
  if (LabelTable[Adr and ROMEnd] = '          ') and PrefAutoLabel then begin
    // Label was empty, so we can add the label
    // First align string to 10 characters
    L := Length(Lab);
    S := '';
    if L > 10 then
      for i := 1 to 10 do S := S + Lab[i]
    else if L < 10 then begin
      S := Lab;
      for i := L to 10 do S := S + ' ';
    end else
      // Length is exactly 10
      S := Lab;
    LabelTable[Adr and ROMEnd] := S;
  end;
end;


procedure FindMainFrameFile(Adr: word; var MainLabel, Comment: string);
// Find MainFrame label from file
var
  S, LabelComment: string;
  CurrentPg: Char;
  Found: boolean;
  LabelAdr, L: integer;
begin
  MainLabel := '';
  Comment := '';
  LabelComment := '';
  CurrentPg := Hex1(Adr shr 12);
  if LabelFileOpen and PrefMainFrame then begin
    // file is open, so start looking ...
    Reset(LabelFile);        // back to beginning
    Found := false;
    while (not EOF(LabelFile)) and (not Found) do begin
      ReadLn(LabelFile, S);
      L := Length(S);
      if (L > 0) and (S[1] = CurrentPg) then begin
        // now we are in business
        LabelAdr := HexToInt(S);
        S := TrimLeft(S);
        L := Length(s);
        if (L > 0) and (S[1] = ':') then begin
          // This is the Marker for the ROM or Section title
          // Delete the : and get the string
          Delete(S, 1, 1);    // remove the :
          LabelComment := Trim(S);
        end else
          if (LabelAdr = Adr) then Found := true;
      end;
    end;
    if Found then begin
      // label address found, now parse Label name and comment
      MainLabel := GetFirstWord(S);   // leading spaces are already removed
      Comment := Trim(S);
      if Comment = '' then Comment := LabelComment;
    end;
  end;
end;


procedure FindXROMFile(I, J: Integer; var S1, S2: string);
  // Finds XROM I, J from own ROM or XROM.TXT
  // Returns: S1: function name, string is empty when not found
  //          S2: comments
var
  S, SS, SSS, XROMComment: string;
  S3: string5;
  S4: string2;
  Found: boolean;
  L, MatchXR: integer;
begin
  S1 := '';
  S2 := '';
  XROMComment := '';
  S3 := Int2(I) + ',' + Int2(J);       // Complete XROM number nn,mm
  S4 := Int2(I);                       // XROM ID only
  if XROMFileOpen and PrefXROM then begin
    Reset(XROMFile);
    Found := false;
    while (not EOF(XROMFile)) and (not Found) do begin
      ReadLn(XROMFile, S);
      L := Length(S);
      SS := GetFirstWord(S);    // read first word from line, could be XROM
      SSS := GetFirstWord(S);   // read 2nd word from line, could be xx,nn
      MatchXR := Pos(S4, SSS);
      if (L > 0) and (SS = 'XROM') and (MatchXR = 1) then begin
        // We are at least in the right XROM
        // First check if we have a comment
        L := Length(S);
        if (L > 0 ) and (S[1] = ':') then begin
          // this is the marker for the XROM name or comment
          Delete(S, 1, 1);  // remove the :
          XROMComment := Trim(S);
        end else if (SSS = S3) then begin
          // This is a real match for the complete XROM number
          S1 := GetFirstWord(S);  // also handle quotes
          S2 := Trim(S);          //this is the comment
          Found := true;
        end;
        if S2 = ''  then S2 := XROMComment;
      end;
    end;
  end;
end;


procedure AddComment(var Line: String; Comment: String);
begin
  if PrefComments and (Comment <> '') then begin
    ExpandSpaces(Line, 54);
    Line := Line + Comment;
  end;
end;


procedure DisAssem(var Adr: word; I: word; var S: string);
// This is the main loop of the Disassembler
type
  TFunPos = (ThisROM, NextROM, PrevROM, OutSide);
var
  Point, j: integer;
  Accumulator: word;
  prompt: word;
  Cmt: String;
  FunPos: TFunPos;

  procedure DecodeClass0(var Adr: word; I: word);
  // Class 0 instructions have the following form :
  //      -- -- -- -- -- -- -- -- 0 0
  //      P3 P2 P1 P0 I3 I2 I1 I0
  //          \/          \/
  //      parameter   instruction
  var
    InstrType, InstrParam: word;
    TempID, NextInstr, ThirdInstr: word;
    SaveAdr: word;
    Comment: String;
    NOPCounter: integer;
  begin
    St := St + Hex3(GetInstr(Adr));

    // if the instruction (mainly NOP) is in the ENTRY table, do not change the ID
    // SetID(Instruction1, Adr);
    InstrType := (I shr 2) and $000F;
    InstrParam := (I shr 6) and $000F;
    if I <> $130 then St := St + EmptyData2;
    case I of
      $000: begin         // this is a NOP
              St := St + Mnemonics[0, PrefMnem + 1];
              // now check if this is the NOP after a function name
              // and if the second instruction is also a NOP
              NextInstr := GetInstr(Adr + 1);
              ThirdInstr := GetInstr(Adr + 2);
              if (Adr <> $0000) and ((Adr and $0FFF) <> $0FFF) then begin
                // Prevents disassembly outside the ROM
                TempID := GetID(Fetch(Adr - 1));
                if TempID = FunctionName then begin
                  Comment := '; NOT programmable and ';
                  if NextInstr = $000 then
                    Comment := Comment + 'Immediate'
                  else
                    Comment := Comment + 'NULLable';
                  ExpandSpaces(St, 37);
                  if PrefComments then St := St + Comment;
                end else if ((Adr and $0FFF) < $0FF4) then begin
                  // This was not a function name but a regular NOP
                  // if there are 3 or more consecutive NOPs without label, skip them
                  if (NextInstr = 0) and (ThirdInstr = 0) and
                     (LabelTable[(Adr + 1) and ROMEnd] = '          ') and
                     (LabelTable[(Adr + 2) and ROMEnd] = '          ') and
                     PrefSkipNOPs then begin
                    // now we have 3 NOPS without label!
                    NOPCounter := 1;
                    SaveAdr := Adr;
                    Adr := Adr + 1;
                    while (GetInstr(Adr) = 0) and
                          (LabelTable[Adr and ROMEnd] = '          ') and
                          (Adr and $0FFF < $0FF4) do begin
                      Inc(Adr);
                      Inc(NOPCounter);
                    end;
                    Adr := Adr - 1;
                    //             8038 - 8057  000 NOP  54 *
                    St := '- ' + Hex4(Adr) + '  000 NOP     ';
                    St := St + '* ' + SStr(NOPCounter) + ' NOPs';
                    Comment := '; ' + SStr(NOPCounter) + ' consecutive NOPs from $';
                    Comment := Comment + Hex4(SaveAdr) + ' to $' + Hex4(Adr);
                    ExpandSpaces(St, 37);
                    if PrefComments then St := St + Comment;
                  end;
                end;
              end;
            end;
      $030: St := St + Mnemonics[1, PrefMnem + 1];
      $040: St := St + Mnemonics[2, PrefMnem + 1];
      $100: St := St + Mnemonics[3, PrefMnem + 1];
      $180: St := St + Mnemonics[4, PrefMnem + 1];
      $140: St := St + Mnemonics[5, PrefMnem + 1];
      $1C0: St := St + Mnemonics[6, PrefMnem + 1];
      $130: begin
          St := St + Space + Hex3(GetInstr(Adr + 1)) + EmptyData;
          St := St + Mnemonics[7, PrefMnem + 1];
          Adr := Adr + 1;
          SetID(Instruction2, Adr);
          St := St + Hex3(GetInstr(Adr));
        end;
      $1F0: St := St + Mnemonics[8, PrefMnem + 1];
      $3C4: St := St + Mnemonics[9, PrefMnem + 1];
      $3C8: St := St + Mnemonics[10, PrefMnem + 1];
      $3CC: St := St + Mnemonics[11, PrefMnem + 1];
      $3D4: St := St + Mnemonics[12, PrefMnem + 1];
      $3DC: St := St + Mnemonics[13, PrefMnem + 1];
    else
      case InstrType of
        $0: St := St + Mnemonics[15, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $1: St := St + Mnemonics[24, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $2: St := St + Mnemonics[25, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $3: St := St + Mnemonics[26, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $4: St := St + Mnemonics[27, PrefMnem + 1] + Hex1(InstrParam);
        $5: St := St + Mnemonics[28, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $6: St := St + Mnemonics[29 + InstrParam, PrefMnem + 1];
        $7: St := St + Mnemonics[45, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $8: St := St + Mnemonics[46 + InstrParam, PrefMnem + 1];
        $9: St := St + Mnemonics[62, PrefMnem + 1] + Hex1(InstrParam);
        $A: St := St + Mnemonics[63, PrefMnem + 1] + ParamDefRegs[InstrParam];
        $B: St := St + Mnemonics[64, PrefMnem + 1] + ParamDefFlag[InstrParam];
        $C: St := St + Mnemonics[65 + InstrParam, PrefMnem + 1];
        $D: St := St + Mnemonics[29, PrefMnem + 1];
        $E: St := St + Mnemonics[81, PrefMnem + 1] + ParamDefRegs[InstrParam];
        $F: St := St + Mnemonics[83, PrefMnem + 1] + ParamDefFlag[InstrParam];
      end;
    end;
  end; // DecodeClass0

  procedure DecodeClass1(var Adr: word; I: word);
  // Class 1 instructions have the following form :
  //      first word:   --- --- --- --- --- --- --- ---  0   1
  //                    A7  A6  A5  A4  A3  A2  A1  A0
  //      second word:  --- --- --- --- --- --- --- --- --- ---
  //                    A15 A14 A13 A12 A11 A10  A9  A8  G   C
  //                                \/                   /   \
  //                           address bits       GOTO/XEQ   condition
  //      G C
  //      0 0   ?NC XQ
  //      0 1   ?C  XQ
  //      1 0   ?NC GO
  //      1 1   ?C  GO
  var
    Dest, SecondWord, SecondID, ThirdWord, Point: word;
    Hulp, Hulp1, Hulp2, Comment: string;
    MSGIdx: integer;
  begin
    // Added to prevent disassembly of non-instruction
    SecondID := GetID(Fetch(Adr + 1));
    SecondWord := GetInstr(Adr + 1);
    Comment := '';
    if (SecondID <> Undefined) and (SecondID <> Instruction2) then begin
      // SecondWord is NOT the second word of an instruction, but something else
      // Current Word will be treated as data
      SetID(Data, Adr);
      St := '';
    end else begin
      Dest := Lo(I shr 2);
      St := St + Hex3(GetInstr(Adr)) + Space + Hex3(GetInstr(Adr + 1));
      Adr := Adr + 1;
      SetID(Instruction2, Adr);
      Hulp := Mnemonics[84 + (SecondWord and $0003), PrefMnem + 1];
      SecondWord := Swap(Lo(SecondWord shr 2));
      Dest := Dest or SecondWord;

      // now see if it is a ?NC XQ to a certain address, so it is in fact
      // a three byte GOTO or GOSUB. If so, decode the third byte also.
      // addresses currently recognised:
      // label   address     action
      // -----   -------     ------
      // GOL0      23D0      relocatable GOTO 1st 1K of ROM
      // GOLSUB0   23D2      relocatable GOSUB 1st 1K of ROM
      // GOL1      23D9      relocatable GOTO 2nd 1K of ROM
      // GOLSUB1   23DB      relocatable GOSUB 2nd 1K of ROM
      // GOL2      23E2      relocatable GOTO 3rd 1K of ROM
      // GOLSUB2   23E4      relocatable GOSUB 3rd 1K of ROM
      // GOL3      23EB      relocatable GOTO 4th 1K of ROM
      // GOLSUB3   23ED      relocatable GOSUB 4th 1K of ROM
      // GOLONG    0FDA      relocatable GOTO same 1K of ROM (or 0FD9, GOLNGH)
      // GOSUB     0FDD      relocatable GOSUB same 1K of ROM (or 0FDE, GOSUBH)
      //
      // MESSL     07EF      left shift the following string into the display
      //                     from the right. The last character in the string
      //                     has $200 added to the character value
      // ?NC XQ    22F5      error display routine ERROR
      // ?NC XQ    1C6B      message display routine MSG
      // ?NC XQ    1C6C      message display routine MSGA

      if Dest = $23D0 then begin              // relocatable GOTO 1st 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[89, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 1st Quad';
      end else if Dest = $23D2 then begin     // relocatable GOSUB 1st 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr)  + CurrentBase;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[88, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 1st Quad';
      end else if Dest = $23D9 then begin     // relocatable GOTO 2nd 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0400;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[89, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 2nd Quad';
      end else if Dest = $23DB then begin     // relocatable GOSUB 2nd 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0400;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[88, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 2nd Quad';
      end else if Dest = $23E2 then begin     // relocatable GOTO 3rd 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0800;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[89, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 3rd Quad';
      end else if Dest = $23E4 then begin     // relocatable GOSUB 3rd 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0800;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[88, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 3rd Quad';
      end else if Dest = $23EB then begin     // relocatable GOTO 4th 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0C00;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[89, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 4th Quad';
      end else if Dest = $23ED then begin     // relocatable GOSUB 4th 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + $0C00;
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[88, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in 4th Quad';
      end else if ((Dest = $0FDA) or (Dest = $0FD9)) then begin
        // relocatable GOTO same 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + (Adr and $0C00);
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[89, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in same Quad';
      end else if ((Dest = $0FDE) or (Dest = $0FDD)) then begin
        // relocatable GOSUB same 1K of ROM
        Adr := Adr + 1;
        St := St + Space + Hex3(GetInstr(Adr));
        SetID(Instruction3, Adr);
        ThirdWord := GetInstr(Adr) + CurrentBase + (Adr and $0C00);
        AddLabel(ThirdWord, 'LB_' + Hex4(ThirdWord));
        Hulp2 := FindLabel(ThirdWord);
        Hulp2 := LeftStr(Hulp2, 7);
        St := St + ' ' + Mnemonics[88, PrefMnem + 1] + Hulp2;
        St := St + ' [' + Hex4(ThirdWord) + ']';
        Comment := '; ' + Hulp + Hex4(Dest) + ', address in same Quad';
      end else if Dest = $07EF then begin
        // left shift the following string into the display
        St := St + EmptyData + Mnemonics[84, PrefMnem + 1] + 'MESL "';
        Point := Adr;
        Hulp := '';
        repeat
          Point := Point + 1;
          SetID(StrDat, Point);
          St := St + HPChar(GetInstr(Point));
          Hulp := Hulp + Hex3(GetInstr(Point)) + ' ';
        until ((GetInstr(Point) > $00FF) or (GetInstr(Point) = $0000));
        // Also stop when there is $000 in the string,
        St := St + '"';
        Comment := '; ' + Hulp;
        Adr := Point;
      end else if (Dest = $22F5) or (Dest = $1C6B) or (Dest = $1C6C) then begin
        // ?NC XQ    22F5      error display routine ERROR
        // ?NC XQ    1C6B      message display routine MSG
        // ?NC XQ    1C6C      message display routine MSGA
        //   MSG_table: array[0..12, 0..2] of string12 = (

        Adr := Adr + 1;
        ThirdWord := GetInstr(Adr);
        Hulp1 := Mnemonics[84, PrefMnem + 1];
        St := St + Space + Hex3(ThirdWord) + Space + Hulp1;
        Hulp1 := Hulp1 + Hex4(Dest);
        SetID(Instruction3, Adr);
        case Dest of
          $22F5: Hulp := 'ERROR   ';
          $1C6B: Hulp := 'MSG     ';
          $1C6C: Hulp := 'MSGA    ';
          else FindMainFrameFile(Dest, Hulp1, Comment);  // should never happen
        end;
        case ThirdWord of
          $018: MSGIdx :=  0; // 'MSGAD ', 'ALPHA DATA'
          $022: MSGIdx :=  1; // 'MSGDE ', 'DATA ERROR'),
          $02D: MSGIdx :=  2; // 'MSGML ', 'MEMORY LOST'),
          $038: MSGIdx :=  3; // 'MSGNE ', 'NONEXISTENT'),
          $03C: MSGIdx :=  4; // 'MSGNL ', 'NULL'),
          $043: MSGIdx :=  5; // 'MSGPR ', 'PRIVATE'),
          $04F: MSGIdx :=  6; // 'MSGOF ', 'OUT OF RANGE'),
          $056: MSGIdx :=  7; // 'MSGWR ', 'PACKING'),
          $05F: MSGIdx :=  8; // 'MSGTA ', 'TRY AGAIN'),
          $062: MSGIdx :=  9; // 'MSGYES', 'YES'),
          $064: MSGIdx := 10; // 'MSGNO ', 'NO'),
          $067: MSGIdx := 11; // 'MSGRAM', 'RAM'),
          $06A: MSGIdx := 12; // 'MSGROM', 'ROM'));
          else  MSGIdx := 13;  // this is a situation with a wrong argument
        end;
        Comment := '; '+ Hulp1 + ': ';
        if MSGIdx < MSG_maxentries then begin
          Hulp := Hulp + MSG_table[MSGIdx, 1];
          Comment := Comment + 'display message "' + MSG_table[MSGIdx, 2] + '"';
        end else begin
          // MSGIdx = 0, unknown argumnet
          Hulp := Hulp + MSG_table[MSGIdx, 1];
          Comment := Comment + 'unknown argument (see 3rd byte) for ERROR/MSG/MSGA' ;
        end;
        St := St + Hulp;
      end else begin
        // normal GO/XQ, find MainFrame Label
        St := St + EmptyData + Hulp;
        // Hulp := FindMainFrame(Dest);
        FindMainFrameFile(Dest, Hulp1, Comment);
        if Comment <> '' then Comment := '; ' + Comment;

        // Add label to our collection if this in the same page
        // or in the same 8K block
        if (BaseAddress = (Dest and $F000)) then
          AddLabel(Dest, 'LB_' + Hex4(Dest));
        if Dis8K and (BaseAddress = (Dest and $E000)) then
          AddLabel(Dest, 'LB_' + Hex4(Dest));

        Hulp := Hulp1;
        if Hulp = '' then
          // no label found
          St := St + Hex4(Dest)
        else begin
          ExpandSpaces(Hulp, 7);
          St := St + Hulp + '[' + Hex4(Dest) + ']';
        end;
      end;
    end;
    if (Comment <> '') and PrefComments then begin
      ExpandSpaces(St, 55 - 18);
      St := St + Comment;
    end;
  end; // DecodeClass1

  procedure DecodeClass2(I: word);
  // Class 2 instructions have the following form :
  //      -- -- -- -- -- -- -- -- 1 0
  //      I4 I3 I2 I1 I0 F2 F1 F0
  //            \/          \/
  //       instruction     field
  var
    Field, Instruct: word;
  begin
    St := St + Hex3(GetInstr(Adr)) + EmptyData2;
    Field := (I shr 2) and $0007;
    Instruct := (I shr 5) and $001F;
    St := St + Mnemonics[90 + Instruct, PrefMnem + 1] + FieldDef[Field];
  end; // DecodeClass2

  procedure DecodeClass3(Adr: word; I: word);
  var
    AdrI: Integer absolute Adr;
    iI: Integer absolute I;
    Dest: word;
    Hulp: string;
    // Class 3 instructions have the following form :
    //      -- -- -- -- -- -- -- -- 1 1
    //      D6 D5 D4 D3 D2 D1 D0  C
    //               \/           \
    //      signed displacement   condition: 0 - JNC, Jump if No Carry
    //                                       1 - JC,  Jump if Carry
  begin
    St := St + Hex3(GetInstr(Adr)) + EmptyData2;
    if ((I and $0004) = $0000) then
      St := St + Mnemonics[123, PrefMnem + 1]
    else
      St := St + Mnemonics[122, PrefMnem + 1];

    I := I shr 3;
    if ((I and $0040) = $0040) then begin      // the jump distance is negative
      I := not I;
      I := I and $003F;
      I := I + 1;
      Dest := Adr - I;  // Destination Address
      St := St + '-';
    end else begin                                           // jump is forward
      I := I and $003F;
      Dest := Adr + I;  // Destination Address
      St := St + '+';
    end;
    AddLabel((Dest), 'LB_' + Hex4(Dest));   // Add this label if it did not exist
    Hulp := FindLabel(Dest);                // get label
    Hulp := LeftStr(Hulp, 7);
    St := St + Hex2(I) + ' ' + Hulp + ' [' + Hex4(Dest) + ']';
  end;                                                          // DecodeClass3

  function Dis: AnyString;
  // does the normal disassembly of instructions
  begin
    St := '';
    Mode := I shr 10;
    I := I and $03FF;
    IClass := I and $0003;
    case IClass of
      0: DecodeClass0(Adr, I);                            // misc. instructions
      1: DecodeClass1(Adr, I);                         // absolute GOTO/EXECUTE
      2: DecodeClass2(I);                      // arithmetic/logical operations
      3: DecodeClass3(Adr, I);                             // PC relative jumps
    end;                                                  //  case Class of ...
    Dis := St;
  end;                                                              // DisAssem

  function HPUserChar(W: word): Char;
  begin
    HPUserChar := Char(CharS[(W and $007F)]);
  end;

  function DatReg(Dat: word): string;
  begin
    Dat := Dat and $7F;
    if (Dat < 100) then DatReg := Int2(Dat) else DatReg := DatRegT[Dat];
  end;

  procedure FindXROM(I, J: Integer; var S1, S2: string);
  // Finds XROM I, J from own ROM or XROM.TXT
  // Returns: S1: function name, string is empty when not found
  //          S2: comments
  begin
    S1 := '';
    S2 := '';
    if (I = ThisXROM) then begin
      // XROM in this ROM
      S1 := ROMFat[j];
      S2 :='in [' + ROMFat[0] + '], this ROM';
      if Dis8K then S2 := S2 + ', 1st 4K'
    end else if (I = ThisXROM8K) and Dis8K then begin
      // XROM is 2nd 8K
      S1 := ROMFat8K[j];
      S2 := 'in [' + ROMFat8K[0] + '], this ROM, 2nd 4K'
    end else begin
      FindXROMFile(I, J, S1, S2);
      // find XROM name from extrenal file
    end;
  end;


  function UserDis(var Adr: word): string;
  // Disassembler for User Code
  var
    Dat, Dat1: word;
    US, US1: string;

    procedure DisUserCode;
    var
      I, j, k: Integer;
      S1, S2, S3, Lbl: string;
      St_Adr: word;
    begin
      Dat := GetInstr(Adr) and $FF;
      Lbl := FindLabel(Adr);
      // if the label starts with LB_, it is an MCode label, should not be here!
      if Pos('LB_', Lbl) = 1 then Lbl := '            ';
      US1 := Lbl + Hex4(Adr) + Space;
      US1 := US1 + Hex3(GetInstr(Adr));
      US := Int3(LinNum) + Space;
      // Code added to put * before label, like in printer listings
      if (GetInstr(Adr) and $00FF) in [$01..$0F, $C0..$CD, $CF]  then begin
        US := US + '*';
      end else
        US := US + Space;
      US := US + HexTable[Dat];
      case TypTable[Dat] of
        OneByte: begin
            SetID(UserProgram, Adr);
            US1 := US1 + EmptyData2;
          end;
        Number: begin
            // Added to allow number entries to be put on one line
            // note that first digit is already there!
            SetID(UserProgram, Adr);
            US1 := US1 + EmptyData2;
            while ((GetInstr(Adr + 1) and $100) = $000) do begin
              // we are not at the end of the number
              Adr := Adr + 1;
              SetID(UserProgram, Adr);
              US := US + HexTable[GetInstr(Adr) and $00FF];
            end;
          end;
        TwoByte: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr)) + EmptyData;
            Dat := GetInstr(Adr) and $FF;
            if Dat < $80 then US := US + DatReg(Dat)
            else US := US + 'IND ' + DatReg(Dat);
            SetID(UserProgram, Adr);
          end;
        XROM: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr)) + EmptyData;
            Dat1 := GetInstr(Adr) and $FF;
            I := ((Dat - $A0) * 4) + (Dat1 div $40);
            j := Dat1 and $3F;
            FindXROM(I, j, S1, S2);
            if S1 <> '' then begin
              US := US + S1 ;
              if PrefComments then begin
                for k := 15 downto Length(S1) do US := US + Space;
                US := US +'; XROM ' + Int2(I) + ',' + Int2(j) + ' ' + S2;
              end;
            end else begin
              // no valid XROM name found
              US := US + Int2(I) + ',' + Int2(j);
              if PrefComments then begin
                for k := 1 to 11 do US := US + Space;
                US := US +'; no XROM name found';
              end;
            end;
            SetID(UserProgram, Adr);
          end;
        GTO: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr)) + EmptyData;
            SetID(UserProgram, Adr);
          end;
        GLOBAL: begin
            SetID(UserProgram, Adr); // 192 = hex code
            St_Adr := Adr;
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr));
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr));
            SetID(UserProgram, Adr); // 241+n=length
            Hulp := '';
            if ((GetInstr(Adr) and $00F0) <> $00F0) then begin
              // final end of usercode program
              Delete(US, Pos('*LBL  "', US), 7);
              US := US + ' END';
              EndUs := True;
              S3 := 'END';
              Hulp := 'Program END';
            end else begin
              // Real label
              j := (GetInstr(Adr) and $0F); // number of chars in string
              Adr := Adr + 1;
              SetID(UserProgram, Adr); // key assigned
              S3 := '';
              for I := 1 to j - 1 do begin
                Adr := Adr + 1;
                SetID(UserProgram, Adr);
                S3 := S3 + HPUserChar(GetInstr(Adr));
                Hulp := Hulp + Hex2(GetInstr(Adr) and $FF) + ' ';
              end;
              AddLabel(St_Adr, 'UC:' + S3);
              US := US + S3 + '"';
            end;
            if PrefComments then begin
              ExpandSpaces(US, 25);
              US := US + '; ' + Hulp;
            end;
            US1 := US1 + Space;
          end;
        ThreeByte: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr));
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr)) + Space;
            SetID(UserProgram, Adr);
            Dat1 := GetInstr(Adr) and $FF;
            US := US + DatReg(Dat1 and $7F);
          end;
        Text: begin
            SetID(UserProgram, Adr);
            Hulp := '';
            US1 := US1 + EmptyData2;
            if ((Dat and $000F) > 0) then begin
              for I := 1 to (Dat and $000F) do begin
                Adr := Adr + 1;
                SetID(UserProgram, Adr);
                US := US + HPUserChar(GetInstr(Adr));
                Hulp := Hulp + Hex2(GetInstr(Adr) and $FF) + ' ';
              end;
              US := US + '"';
            end;
            if PrefComments then begin
              ExpandSpaces(US, 25);
              US := US + '; ';
              if (Dat and $000F) = 0 then
                US := US + 'zero-length text, 1-byte NOP'
              else
                US := US + Hulp;
            end;
          end;
        GotoText: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            SetID(UserProgram, Adr);
            Dat := GetInstr(Adr) and $FF;
            Hulp := '';
            US1 := US1 + EmptyData2;
            if ((Dat and $000F) > 0) then
              for I := 1 to (Dat and $000F) do begin
                Adr := Adr + 1;
                SetID(UserProgram, Adr);
                US := US + HPUserChar(GetInstr(Adr));
                Hulp := Hulp + Hex2(GetInstr(Adr) and $FF) + ' ';
              end;
            US := US + '"';
            if PrefComments then begin
              ExpandSpaces(US, 25);
              US := US + '; ' + Hulp;
            end;
          end;
        GXIND: begin
            SetID(UserProgram, Adr);
            Adr := Adr + 1;
            US1 := US1 + Space + Hex3(GetInstr(Adr)) + EmptyData;
            SetID(UserProgram, Adr);
            Dat := GetInstr(Adr) and $FF;
            if Dat > $7F then US := US + 'XEQ IND '
            else US := US + 'GTO IND ';
            US := US + DatReg(Dat and $7F);
          end;
      end;
      LinNum := LinNum + 1;
      US1 := US1 + US;
      US := US1;
    end; // DisUserCode

  begin {UserDis }
    LinNum := 1;
    EndUs := False;
    repeat
      DisUserCode;
      Adr := Adr + 1;
      if DisasmPass = 2 then                    // only add to putput on Pass 2!
        ListWindow.FrmLister.EdtList.Lines.Add(US);
    until EndUs;
    Adr := Adr - 1;
    UserDis := '            END OF USER CODE';
    AddEmptyLine := true;
  end; { UserDis }

begin { DisAssem }
  { the disassembler starts here. Parameters given:
    VAR Adr : word       the address of the data to be disassembled
          i : word       the data at the current address
    VAR   S : String80   the target string for the final result
    The specification for this string is :
     LABEL   - 10 positions
     ADDRESS - 4 positions
     DATA    - three times three words
     and then the instruction, or explanation of data.
     In case of a data string, all words are on the line.
     Note that some labels are predefined                            }
  case GetID(I) of

    Undefined: { this word should be a normal instruction }
      begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        Hulp := Dis;
        if Hulp = '' then S := '' else S := S + Hulp;
      end;

    XROMNumber: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;
        AddComment(S, '; XROM number    %' + Int3(GetInstr(Adr)));
        if In8K then
          ThisXROM8K := GetInstr(Adr)
        else
          ThisXROM := GetInstr(Adr);
      end;

    FATEntries: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;
        AddComment(S, '; # FAT Entries  %' + Int3(GetInstr(Adr)));
        AddEmptyLine := true;
      end;

    FATFunction: { find the function name and analyse it }
      begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + Space;
        S := S + Hex3(GetInstr(Adr + 1)) + EmptyData;
        FirstWord := GetInstr(Adr);
        Adr := Adr + 1;
        SecondWord := GetInstr(Adr);
        StartAdr := (SecondWord and $FF);
        StartAdr := StartAdr or ((FirstWord and $FF) shl 8);
        Cmt := '';

        // StartAdr can be one of the following:
        //     0xxx - Function Name in this 4K block
        //            - This is always valid
        //     1xxx - Function Name in next 4K block
        //            - valid only when we are doing the lower 4K
        //     Fxxx - Function Name in previous 4K block
        //            - valid only when we are doing the upper 4K

        // One of the following siutations may be true:
        //          disassembly in Page   Function in Page    StartAddr
        //     A -       lower 4K             lower 4K          0xxx
        //     B -       lower 4K             upper 4K          1xxx
        //     C -       upper 4K             lower 4K          Fxxx
        //     D -       upper 4K             lower 4K          0xxx
        //     E -       Function Name out of reach

        // First find out if we have a valid StartAdr for which we can decode the name
        if (StartAdr and $F000) = $0000 then begin
          // We are in our own Page
          FunPos := ThisROM;
          Point := (StartAdr and $0FFF) + CurrentBase;
          S := S + Hex4(Point) + Space2;
        end else if Dis8K and ((StartAdr and $F000) = $1000) and (not in8K) then begin
          // we are in the lower 4K (even page) and the Function Name is in the upper 8K
          FunPos := NextROM;
          Point :=  (StartAdr and $0FFF) + BaseAddress + $1000;
          S := S + Hex4(Point) + Space2;
          Cmt := '; Function in Upper 4K';
        end else if Dis8K and ((StartAdr and $F000) = $F000) and in8K then begin
          // we are in the upper 4K (odd page) and the Function Name is in the lower 4K
          FunPos := NextROM;
          Point :=  (StartAdr and $0FFF) + BaseAddress;
          S := S + Hex4(Point) + Space2;
          Cmt := '; Function in Lower 4K';
        end else begin
          // we are outside our own Page, and cannot access the other page
          FunPos := OutSide;
          Point := StartAdr;
          S := S + Hex4(Point) + Space2;
          Cmt := '; Function outside page ';
          if ((FirstWord and $0200) = $0200) then
            //Usercode program
            AddComment(S, Cmt + ' (UCode)')
          else
            AddComment(S, Cmt + ' (MCode)');
        end;

        StartAdr := Point;
        if FunPos <> OutSide then begin
          // We have a valid Function Name address in Point
          // First sort out the function name
          if ((FirstWord and $0200) = $0200) then begin
            // This is a UserCode program
            Point := StartAdr + 2;
            LabLen := GetInstr(Point) and $000F;
            Hulp := '';
            for j := 2 to LabLen do begin
              W := GetInstr(Point + j) and $00FF;
              Hulp := Hulp + HPUserChar(W);
            end;
            S := S + '"' + Hulp + '"';
            for j := LabLen to 10 do S := S + ' ';
            if Cmt = '' then Cmt := ';';
            AddComment(S, Cmt + ' (UCode)');
            // Now put the Label in the right location, name in Hulp
            AddLabel(Point - 2, 'U:' + Hulp);
            // added to allow usercode before the Global Label to be disassembled
            j := 0;
            repeat
              j := j + 1;
              WW := GetInstr(StartAdr - j) and $0200;
              if WW <> $0200 then begin
                SetID(UserProgram, StartAdr - j);
                ClearMCodeLabel(StartAdr-1);
              end;
            until (WW = $0200);
            SetID(FirstTwo, StartAdr - j - 1);
            SetID(FirstTwo, StartAdr - j);
            for Point := StartAdr to StartAdr + 3 do begin
              SetID(UserProgram, Point);
              ClearMCodeLabel(Point);
            end;
          end else begin
            // This is an MCode program
            Point := StartAdr;
            // next statement inserted for correct disassembly of 8K roms
            if GetID(Fetch(Point)) <> FunctionName then
              SetID(Instruction1, Point);
            Hulp := '';
            repeat
              Point := Point - 1;
              W := GetInstr(Point);
              Hulp := Hulp + HPChar(W);
              SetID(FunctionName, Point);
            until ((GetInstr(Point) and $0080) <> 0) or (Point = $0000);
            S := S + Hulp;
            AddLabel(StartAdr, '"' + Hulp + '"');
            AddComment(S, Cmt);
          end;
          // The Function Name is now handled, only add to the correct XROM table
          if not in8K then
            ROMFat[(((Adr and $0FFF)- 1) DIV 2) - 1] := Hulp    // In lower page
          else
            ROMFat8K[(((Adr and $0FFF)- 1) DIV 2) - 1] := Hulp; // Points to upper page
        end;
      end;

    FATEnd: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + Space;
        S := S + Hex3(GetInstr(Adr + 1)) + EmptyData;
        AddComment(S, '; End of FAT');
        Adr := Adr + 1;
        AddEmptyLine := true;
      end;

    FunctionName: begin
        A := Adr;
        repeat
          A := A + 1;
        until GetID(Fetch(A)) <> FunctionName;
        // this fails when function names are immediatley following each other
        // See -PRINTER- entry in PRinter ROM 2E, not in IL Printer
        // will consider this for a later change, search in FAT??

        // Added to insert an empty line before the start of a new function
        if DisasmPass = 2 then
          // Only add to output on Pass 2
          ListWindow.FrmLister.EdtList.Lines.Add('');
        S := FindLabel(Adr) + Hex4(A) + EmptyData3;
        S := S + '"';
        Hulp := '; ';

        for Point := A - 1 downto Adr do begin
          S := S + HPChar(GetInstr(Point));
          Hulp := Hulp + Hex3(GetInstr(Point)) + ' ';
        end;
        S := S + '"';
        Adr := A - 1;   // Adr now contains the first character of the function name
        prompt := (GetInstr(Adr) and $0300) shr 8;

        if (GetInstr(Adr) and $0080) = $0000 then
          // Function name was > 1 character, normal case
          prompt := prompt or ((GetInstr(Adr - 1) and $0300) shr 4);
        // else Function name was 1 character, can this happen?
        case prompt of
          $00 : Cmt := '';            // no prompting
          $01 : Cmt := 'Prompt: Alpha (null input valid)';
          $02 : Cmt := 'Prompt: 2 digit ST INF IND_ST + - * /';
          $03 : Cmt := 'Prompt: 2 digit or non-null Alpha';
          $11 : Cmt := 'Prompt: 3 digits';
          $12 : Cmt := 'Prompt: 2 digits ST IND IND_ST';
          $13 : Cmt := 'Prompt: 2 digits IND IND_ST non-null Alpha';
          $21 : Cmt := 'Prompt: non-null Alpha';
          $22 : Cmt := 'Prompt: 2 digits IND IND_ST';
          $23 : Cmt := 'Prompt: 2 digits or non-null Alpha';
          $31 : Cmt := 'Prompt: 1 digit IND IND_ST';
          $32 : Cmt := 'Prompt: 2 digits IND IND_ST';
          $33 : Cmt := 'Prompt: 2 digits IND IND_ST non-null Alpha . ..';
          else  Cmt := 'Prompt: Invalid combination';
        end;
        if Cmt <> '' then Cmt := '/ ' + Cmt;

        AddComment(S, Hulp + Cmt);
      end;

    Data: begin
        if PrefAutoLabel then
          S := 'DATA        '
        else
          S := '            ';
        S := S + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;
        S := S + '"' + HPChar(GetInstr(Adr)) + '"';
      end;

    Instruction1: begin { normal instruction disassembly }
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Dis;
      end;

    Instruction2: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space + 'Instruction2';
      end;

    Instruction3: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space + 'Instruction3';
      end;

    UserProgram: begin
        S := UserDis(Adr);
      end;

    EntryPoint: begin
        if (DisasmPass = 2) and ((Adr and $0FFF) = $0FF4) then
          // new line before start of entry tabel
          ListWindow.FrmLister.EdtList.Lines.Add('');
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Dis;
        if PrefComments then begin
          ExpandSpaces(S, 54);
          case (Adr and $0FFF) of
            $0FF4: S := S + '; PAUSE interrupt';
            $0FF5: S := S + '; After each program line';
            $0FF6: S := S + '; Awaken from DEEP SLEEP';
            $0FF7: S := S + '; Turn OFF by key or XEQ OFF';
            $0FF8: S := S + '; I/O service flag set';
            $0FF9: S := S + '; Awaken by pressing ON key';
            $0FFA: S := S + '; MEMORY LOST interrupt';
          else ;
          end;
        end;
      end;

    ROMRevision: begin
        if DisasmPass = 2 then
          ListWindow.FrmLister.EdtList.Lines.Add('');
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;
        S := S + '"' + HPChar(GetInstr(Adr + 3)) + HPChar(GetInstr(Adr + 2));
        S := S + '-' + HPChar(GetInstr(Adr + 1)) + HPChar(GetInstr(Adr)) + '"    ';
        Hulp := '; ';
        for j := 3 downto 0 do
          Hulp := Hulp + Hex3(GetInstr(Adr + j)) + ' ';
        AddComment(S, Hulp);
        Adr := Adr + 3;
      end;

    CheckSum: begin
        if DisasmPass = 2 then
          ListWindow.FrmLister.EdtList.Lines.Add('');
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;

        // Calculate checksum and compare
        Accumulator := 0;
        if in8K then begin
          for i := $1000 to $1FFE do begin
            Accumulator := Accumulator + GetInstr(i);
            if Accumulator > $03FF then
              Accumulator := (Accumulator and $03FF) + 1;
          end;
        end else begin
          // We are in 1st 4K block
          for i := $0000 to $0FFE do begin
            Accumulator := Accumulator + GetInstr(i);
            if Accumulator > $03FF then
              Accumulator := (Accumulator and $03FF) + 1;
          end;
        end;
        Accumulator := (-Accumulator) and $03FF;
        AddComment(S, '; Calculated Checksum: ' + Hex3(Accumulator));
        if not in8K and (DisasmPass = 2) then AddEmptyLine := true;
      end;

    FirstTwo: begin
        // Added to insert an empty line before the start of Usercode
        if DisasmPass = 2 then
          ListWindow.FrmLister.EdtList.Lines.Add('');

        // if the label starts with LB_, it is an MCode label, should not be here!
        Hulp := FindLabel(Adr);
        if Pos('LB_', Hulp) = 1 then Hulp := '            ';
        S := Hulp + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + Space;
        S := S + Hex3(GetInstr(Adr + 1)) + EmptyData;

        // Added to explain meaning of first two bytes
        Hulp := '; UserCode: ';
        Hulp := Hulp + SStr(GetInstr(Adr) * 7 + ((GetInstr(Adr + 1) and $00F0) shr 4));
        Hulp := Hulp + ' bytes (';
        Hulp := Hulp + SStr(GetInstr(Adr)) + ' regs + ';
        Hulp := Hulp + SStr((GetInstr(Adr + 1) and $00F0) shr 4) + ' bytes) ';
        if ((GetInstr(Adr + 1) and $0100) = $0100) then
          Hulp := Hulp + 'PRIVATE'
        else
          Hulp := Hulp + 'NONPRIVATE';
        AddComment(S, Hulp);
        Adr := Adr + 1;
      end;

    StrDat: begin
        S := FindLabel(Adr) + Hex4(Adr) + Space;
        S := S + Hex3(GetInstr(Adr)) + EmptyData2;
        S := S + '"' + HPChar(GetInstr(Adr)) + '"';
      end;

    else begin
        // unknown ID
        S := 'UNKNOWN CODE';

    end;
  end; { Case }
end;

end. 

